"""Metrics for cross validation with Gaussian graphical models.
"""
import numpy as np
from sklearn.utils.extmath import fast_logdet


def log_likelihood(covariance, precision):
    """Computes the log-likelihood between the covariance and precision 
    estimate.
    
    Parameters
    ----------
    covariance : 2D ndarray (n_features, n_features)
        Maximum Likelihood Estimator of covariance
    
    precision : 2D ndarray (n_features, n_features)
        The precision matrix of the covariance model to be tested
    
    Returns
    -------
    log-likelihood
    """
    assert covariance.shape == precision.shape
    dim, _ = precision.shape
    log_likelihood_ = -np.sum(covariance * precision) +\
                      fast_logdet(precision) - dim * np.log(2 * np.pi)
    log_likelihood_ /= 2.
    return log_likelihood_


def kl_loss(covariance, precision):
    """Computes the KL divergence between precision estimate and 
    reference covariance.
    
    The loss is computed as:

        Trace(Theta_1 * Sigma_0) - log(Theta_0 * Sigma_1) - dim(Sigma)

    Parameters
    ----------
    covariance : 2D ndarray (n_features, n_features)
        Maximum Likelihood Estimator of covariance
    
    precision : 2D ndarray (n_features, n_features)
        The precision matrix of the covariance model to be tested
    
    Returns
    -------
    KL-divergence 
    """
    assert covariance.shape == precision.shape
    dim, _ = precision.shape 
    logdet_p_dot_c = fast_logdet(np.dot(precision, covariance))
    return 0.5 * (np.sum(precision * covariance) - logdet_p_dot_c - dim)


def quadratic_loss(covariance, precision):
    """Computes ...
    
    Parameters
    ----------
    covariance : 2D ndarray (n_features, n_features)
        Maximum Likelihood Estimator of covariance
    
    precision : 2D ndarray (n_features, n_features)
        The precision matrix of the model to be tested
    
    Returns
    -------
    Quadratic loss
    """
    assert covariance.shape == precision.shape
    dim, _ = precision.shape
    return np.trace((np.dot(covariance, precision) - np.eye(dim))**2)


def ebic(covariance, precision, n_samples, n_features, gamma=0):
    '''
    Extended Bayesian Information Criteria for model selection.

    When using path mode, use this as an alternative to cross-validation for
    finding lambda.

    See:
        "Extended Bayesian Information Criteria for Gaussian Graphical Models"
        R. Foygel and M. Drton, NIPS 2010

    Parameters
    ----------
    covariance : 2D ndarray (n_features, n_features)
        Maximum Likelihood Estimator of covariance (sample covariance)
    
    precision : 2D ndarray (n_features, n_features)
        The precision matrix of the model to be tested

    n_samples :  int
        Number of examples.

    n_features : int
        Dimension of an example.

    lam: (float)
        Threshold value for precision matrix. This should be lambda scaling
        used to obtain this estimate.

    gamma : (float) \in (0, 1)
        Choice of gamma=0 leads to classical BIC
        Positive gamma leads to stronger penalization of large graphs.

    Returns
    -------
    ebic score (float).  Caller should minimized this score.
    '''    
    l_theta = -np.sum(covariance * precision) + fast_logdet(precision)
    l_theta *= n_features / 2.

    mask = np.abs(precision.flat) > np.finfo(precision.dtype).eps
    precision_nnz = (np.sum(mask) - n_features) / 2.0  # lower off diagonal triangle

    return -2.0 * l_theta +\
            precision_nnz * np.log(n_samples) +\
            4.0 * precision_nnz * np.log(n_features) * gamma



import time
import collections
import operator
import numpy as np

from sklearn.covariance import EmpiricalCovariance
from sklearn.utils import check_array, as_float_array
from sklearn.utils.testing import assert_array_almost_equal
from sklearn.externals.joblib import Parallel, delayed
#from sklearn.model_selection import check_cv, cross_val_score # >= 0.18
from sklearn.cross_validation import check_cv, cross_val_score # < 0.18

from . import pyquic
from .inverse_covariance import (
    InverseCovarianceEstimator,
    _init_coefs,
    _compute_error,
    _validate_path,
)


def quic(S, lam, mode='default', tol=1e-6, max_iter=1000, 
        Theta0=None, Sigma0=None, path=None, msg=0):
    """Fits the inverse covariance model according to the given training 
    data and parameters.

    Parameters
    -----------
    S : 2D ndarray, shape (n_features, n_features)
        Empirical covariance or correlation matrix.

    Other parameters described in `class InverseCovariance`.

    Returns
    -------
    Theta : 
    Sigma : 
    opt : 
    cputime : 
    iters : 
    dGap : 
    """
    assert mode in ['default', 'path', 'trace'],\
            'mode = \'default\', \'path\' or \'trace\'.'

    Sn, Sm = S.shape
    if Sn != Sm:
        raise ValueError("Input data must be square. S shape = {}".format(
                         S.shape))
        return

    # Regularization parameter matrix L.
    if isinstance(lam, float):
        _lam = np.empty((Sn, Sm))
        _lam[:] = lam
        _lam[np.diag_indices(Sn)] = 0. # make sure diagonal is zero
    else:
        assert lam.shape == S.shape, 'lam, S shape mismatch.'
        _lam = as_float_array(lam, copy=False, force_all_finite=False)
 
    # Defaults.
    optSize = 1
    iterSize = 1
    if mode is "trace":
        optSize = max_iter

    # Default Theta0, Sigma0 when both are None.
    if Theta0 is None and Sigma0 is None:
        Theta0 = np.eye(Sn)
        Sigma0 = np.eye(Sn)

    assert Theta0 is not None,\
            'Theta0 and Sigma0 must both be None or both specified.'
    assert Sigma0 is not None,\
            'Theta0 and Sigma0 must both be None or both specified.'
    assert Theta0.shape == S.shape, 'Theta0, S shape mismatch.'
    assert Sigma0.shape == S.shape, 'Theta0, Sigma0 shape mismatch.'
    Theta0 = as_float_array(Theta0, copy=False, force_all_finite=False)
    Sigma0 = as_float_array(Sigma0, copy=False, force_all_finite=False)

    if mode == 'path':
        assert path is not None, 'Please specify the path scaling values.'

        # path must be sorted from largest to smallest and have unique values
        check_path = sorted(set(path), reverse=True)
        assert_array_almost_equal(check_path, path)

        path_len = len(path)
        optSize = path_len
        iterSize = path_len

        # Note here: memory layout is important:
        # a row of X/W holds a flattened Sn x Sn matrix,
        # one row for every element in _path_.
        Theta = np.empty((path_len, Sn * Sn))
        Theta[0,:] = Theta0.ravel()
        Sigma = np.empty((path_len, Sn * Sn))
        Sigma[0,:] = Sigma0.ravel()
    else:
        path = np.empty(1)
        path_len = len(path)

        Theta = np.empty(Theta0.shape)
        Theta[:] = Theta0
        Sigma = np.empty(Sigma0.shape)
        Sigma[:] = Sigma0
                    
    # Run QUIC.
    opt = np.zeros(optSize)
    cputime = np.zeros(optSize)
    dGap = np.zeros(optSize)
    iters = np.zeros(iterSize, dtype=np.uint32)
    pyquic.quic(mode, Sn, S, _lam, path_len, path, tol, msg, max_iter,
                Theta, Sigma, opt, cputime, iters, dGap)

    if optSize == 1:
        opt = opt[0]
        cputime = cputime[0]
        dGap = dGap[0]

    if iterSize == 1:
        iters = iters[0]

    # reshape Theta, Sigma in path mode
    Theta_out = Theta 
    Sigma_out = Sigma
    if mode == 'path':
        Theta_out = []
        Sigma_out = []
        for lidx in range(path_len):
            Theta_out.append(np.reshape(Theta[lidx, :], (Sn, Sn)))
            Sigma_out.append(np.reshape(Sigma[lidx, :], (Sn, Sn)))

    return Theta_out, Sigma_out, opt, cputime, iters, dGap


class QuicGraphLasso(InverseCovarianceEstimator):
    """
    Computes a sparse inverse covariance matrix estimation using quadratic 
    approximation. 

    The inverse covariance is estimated the sample covariance estimate 
    $S$ as an input such that: 

    $T_hat = max_{\Theta} logdet(Theta) - Trace(ThetaS) - \lambda|\Theta|_1 $

    Parameters
    -----------        
    lam : scalar or 2D ndarray, shape (n_features, n_features) (default=0.5)
        Regularization parameters per element of the inverse covariance matrix.
        
        If a scalar lambda is used, a penalty matrix will be generated 
        containing lambda for all values in both upper and lower triangles
        and zeros along the diagonal.  This differs from the scalar graphical 
        lasso by the diagonal. To replicate the scalar formulation you must 
        manualy pass in lam * np.ones((n_features, n_features)).
    
    mode : one of 'default', 'path', or 'trace'
        Computation mode.

    tol : float (default=1e-6)
        Convergence threshold.

    max_iter : int (default=1000)
        Maximum number of Newton iterations.

    Theta0 : 2D ndarray, shape (n_features, n_features) (default=None) 
        Initial guess for the inverse covariance matrix. If not provided, the 
        diagonal identity matrix is used.

    Sigma0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the covariance matrix. If not provided the diagonal 
        identity matrix is used.

    path : array of floats (default=None)
        In "path" mode, an array of float values for scaling lam.
        The path must be sorted largest to smallest.  This class will auto sort
        this, in which case indices correspond to self.path_

    method : 'quic' or 'bigquic', ... (default=quic)
        Currently only 'quic' is supported.

    verbose : integer
        Used in quic routine.

    score_metric : one of 'log_likelihood' (default), 'frobenius', 'spectral',
                  'kl', or 'quadratic'
        Used for computing self.score().

    init_method : one of 'corrcoef', 'cov', or custom function
        Computes initial covariance and scales lambda appropriately.
        Custom function must return ((n_features, n_features) ndarray, float)
        where the scalar parameter will be used to scale the penalty lam.

    auto_scale : bool
        If True, will compute self.lam_scale_ = max off-diagonal value when 
        init_method='cov'. 
        If false, then self.lam_scale_ = 1.
        lam_scale_ is used to scale user-supplied self.lam during fit.

    Methods
    ----------
    lam_at_index(lidx) :  Compute the scaled lambda used at index lidx.
        The parameter lidx is ignored when mode='default'.  Can use self.lam_
        for convenience in this case.

    Attributes
    ----------
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix
        If mode='path', this is 2D ndarray, shape (len(path), n_features ** 2)

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.
        If mode='path', this is 2D ndarray, shape (len(path), n_features ** 2)

    sample_covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated sample covariance matrix

    lam_ : (float) or 2D ndarray, shape (n_features, n_features)
        When mode='default', this is the lambda used in fit (lam * lam_scale_)

    lam_scale_ : (float)
        Additional scaling factor on lambda (due to magnitude of 
        sample_covariance_ values).

    path_ : None or array of floats
        Sorted (largest to smallest) path.  This will be None if not in path
        mode.

    opt_ :

    cputime_ :

    iters_ :    

    duality_gap_ :

    """
    def __init__(self, lam=0.5, mode='default', tol=1e-6, max_iter=1000,
                 Theta0=None, Sigma0=None, path=None, method='quic', verbose=0,
                 score_metric='log_likelihood', init_method='corrcoef',
                 auto_scale=True):
        # quic-specific params
        self.lam = lam
        self.mode = mode
        self.tol = tol
        self.max_iter = max_iter
        self.Theta0 = Theta0
        self.Sigma0 = Sigma0
        self.method = method
        self.verbose = verbose
        self.path = path

        if self.mode == 'path' and path is None:
            raise ValueError("path required in path mode.")
            return

        self.path_ = _validate_path(self.path)

        # quic-specific outputs
        self.opt_ = None
        self.cputime_ = None
        self.iters_ = None
        self.duality_gap_ = None

        # these must be updated upon self.fit()
        self.sample_covariance_ = None
        self.lam_scale_ = None
        self.is_fitted = False

        super(QuicGraphLasso, self).__init__(
                score_metric=score_metric, 
                init_method=init_method,
                auto_scale=auto_scale)

    def fit(self, X, y=None, **fit_params):
        """Fits the inverse covariance model according to the given training 
        data and parameters.

        Parameters
        -----------
        X : 2D ndarray, shape (n_features, n_features)
            Input data.

        Returns
        -------
        self
        """
        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)
        self.init_coefs(X)
        if self.method == 'quic':
            (self.precision_, self.covariance_, self.opt_, self.cputime_, 
            self.iters_, self.duality_gap_) = quic(self.sample_covariance_,
                                                self.lam * self.lam_scale_,
                                                mode=self.mode,
                                                tol=self.tol,
                                                max_iter=self.max_iter,
                                                Theta0=self.Theta0,
                                                Sigma0=self.Sigma0,
                                                path=self.path_,
                                                msg=self.verbose)
        else:
            raise NotImplementedError(
                "Only method='quic' has been implemented.")

        self.is_fitted = True
        return self

    def lam_at_index(self, lidx):
        """Compute the scaled lambda used at index lidx.
        """
        if self.path_ is None:
            return self.lam * self.lam_scale_

        return self.lam * self.lam_scale_ * self.path_[lidx]

    @property
    def lam_(self):
        if self.path_ is not None:
            print ('lam_ is an invalid parameter in path mode, '
                   'use self.lam_at_index')
        return self.lam_at_index(0)
    

def _quic_path(X, path, X_test=None, lam=0.5, tol=1e-6,
         max_iter=1000, Theta0=None, Sigma0=None, method='quic', 
         verbose=0, score_metric='log_likelihood',
         init_method='corrcoef'):
    """Wrapper to compute path for example X.
    """
    S, lam_scale_ = _init_coefs(
            X,
            method=init_method)

    path = path.copy(order='C')
    if method == 'quic':
        (precisions_, covariances_, opt_, cputime_, 
        iters_, duality_gap_) = quic(S,
                                    lam,
                                    mode='path',
                                    tol=tol,
                                    max_iter=max_iter,
                                    Theta0=Theta0,
                                    Sigma0=Sigma0,
                                    path=path,
                                    msg=verbose)
    else:
        raise NotImplementedError(
            "Only method='quic' has been implemented.")

    if X_test is not None:
        S_test, lam_scale_test = _init_coefs(
            X_test,
            method=init_method)
        
        path_errors = []
        for lidx, lam in enumerate(path):
            path_errors.append(_compute_error(S_test,
                                            covariances_[lidx],
                                            precisions_[lidx],
                                            score_metric=score_metric))
        scores_ = [-e for e in path_errors]

        return covariances_, precisions_, scores_
    
    return covariances_, precisions_


class QuicGraphLassoCV(InverseCovarianceEstimator):
    """Sparse inverse covariance w/ cross-validated choice of the l1 penalty
    via quadratic approximation.  

    This takes advantage of "path" mode in QuicGraphLasso.
    See sklearn.covariance.graph_lasso.GraphLassoCV.

    Parameters
    -----------  
    lam : scalar or 2D ndarray, shape (n_features, n_features) (default=0.5)
        Regularization parameters per element of the inverse covariance matrix.
        The next parameter 'lams' scale this matrix as the lasso path is learned.

        If a scalar lambda is used, a penalty matrix will be generated 
        containing lambda for all values in both upper and lower triangles
        and zeros along the diagonal.  This differs from the scalar graphical 
        lasso by the diagonal. To replicate the scalar formulation you must 
        manualy pass in lam * np.ones((n_features, n_features)).

    lams : integer, or list positive float, optional
        If an integer is given, it fixes the number of points on the
        grids of alpha to be used. If a list is given, it gives the
        grid to be used. See the notes in the class docstring for
        more details.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
        
        - None, to use the default 3-fold cross-validation,
        - integer, to specify the number of folds.
        - An object to be used as a cross-validation generator.
        - An iterable yielding train/test splits.
        
        For integer/None inputs :class:`KFold` is used.
        
        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
    
    n_refinements: strictly positive integer
        The number of times the grid is refined. Not used if explicit
        values of alphas are passed.

    n_jobs: int, optional
        number of jobs to run in parallel (default 1).

    tol : float (default=1e-6)
        Convergence threshold.

    max_iter : int (default=1000)
        Maximum number of Newton iterations.

    Theta0 : 2D ndarray, shape (n_features, n_features) (default=None) 
        Initial guess for the inverse covariance matrix. If not provided, the 
        diagonal identity matrix is used.

    Sigma0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the covariance matrix. If not provided the diagonal 
        identity matrix is used.

    method : 'quic' or 'bigquic', ... (default=quic)
        Currently only 'quic' is supported.

    score_metric : one of 'log_likelihood' (default), 'frobenius', 'spectral',
                  'kl', or 'quadratic'
        Used for computing self.score().

    init_method : one of 'corrcoef', 'cov', or custom function
        Computes initial covariance and scales lambda appropriately.
        Custom function must return ((n_features, n_features) ndarray, float)
        where the scalar parameter will be used to scale the penalty lam.

    auto_scale : bool
        If True, will compute self.lam_scale_ = max off-diagonal value when 
        init_method='cov'. 
        If false, then self.lam_scale_ = 1.
        lam_scale_ is used to scale user-supplied self.lam during fit.

    Attributes
    ----------    
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.

    lam_ : float
        Penalization parameter selected.
    
    cv_lams_ : list of float
        All penalization parameters explored.
    
    grid_scores: 2D numpy.ndarray (n_alphas, n_folds)
        Log-likelihood score on left-out data across folds.
    
    n_iter_ : int
        Number of iterations run for the optimal alpha.
    """
    def __init__(self, lam=1.0, lams=4, n_refinements=4, cv=None, tol=1e-6,
                 max_iter=1000, Theta0=None, Sigma0=None, method='quic', 
                 verbose=0, n_jobs=1, score_metric='log_likelihood',
                 init_method='corrcoef', auto_scale=True):
        # GridCV params
        self.n_jobs = n_jobs
        self.cv = cv
        self.lam = lam
        self.lams = lams
        self.n_refinements = n_refinements

        # quic-specific params
        self.tol = tol
        self.max_iter = max_iter
        self.Theta0 = Theta0
        self.Sigma0 = Sigma0
        self.method = method
        self.verbose = verbose

        # quic-specific outputs
        self.opt_ = None
        self.cputime_ = None
        self.iters_ = None
        self.duality_gap_ = None

        # these must be updated upon self.fit()
        self.sample_covariance_ = None
        self.lam_scale_ = None
        self.is_fitted = False

        super(QuicGraphLassoCV, self).__init__(
                score_metric=score_metric,
                init_method=init_method,
                auto_scale=auto_scale)

    def fit(self, X, y=None):
        """Fits the GraphLasso covariance model to X.
        
        Closely follows sklearn.covariance.graph_lasso.GraphLassoCV.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the covariance estimate
        """
        # initialize
        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)
        cv = check_cv(self.cv, X, y, classifier=False)
        self.init_coefs(X)

        # get path
        if isinstance(self.lams, collections.Sequence):
            path = self.lams
            n_refinements = 1
        else:
            n_refinements = self.n_refinements
            lam_1 = self.lam_scale_
            lam_0 = 1e-2 * lam_1
            path = np.logspace(np.log10(lam_0), np.log10(lam_1), self.lams)[::-1]

        # run this thing a bunch
        results = list()
        t0 = time.time()
        for rr in range(n_refinements):
            # parallel version
            this_result = Parallel(
                n_jobs=self.n_jobs,
                verbose=self.verbose,
            )(
                delayed(_quic_path)(
                    X[train],
                    path,
                    X_test=X[test],
                    lam=self.lam, tol=self.tol, max_iter=self.max_iter, 
                    Theta0=self.Theta0, Sigma0=self.Sigma0, method=self.method,
                    verbose=self.verbose, score_metric=self.score_metric,
                    init_method=self.init_method)
                for train, test in cv)

            # Little dance to transform the list in what we need
            covs, _, scores = zip(*this_result)
            covs = zip(*covs)
            scores = zip(*scores)
            results.extend(zip(path, scores, covs))
            results = sorted(results, key=operator.itemgetter(0), reverse=True)

            # Find the maximum (avoid using built in 'max' function to
            # have a fully-reproducible selection of the smallest alpha
            # in case of equality)
            best_score = -np.inf
            last_finite_idx = 0
            for index, (lam, scores, _) in enumerate(results):
                # sometimes we get -np.inf in the result (in kl-loss)
                scores = [s for s in scores if not np.isinf(s)]
                if len(scores) == 0:
                    this_score = -np.inf 
                else:
                    this_score = np.mean(scores)

                if this_score >= .1 / np.finfo(np.float64).eps:
                    this_score = np.nan

                if np.isfinite(this_score):
                    last_finite_idx = index
                
                if this_score >= best_score:
                    best_score = this_score
                    best_index = index

            # Refine the grid
            if best_index == 0:
                # We do not need to go back: we have chosen
                # the highest value of lambda for which there are
                # non-zero coefficients
                lam_1 = results[0][0]
                lam_0 = results[1][0]
            
            elif (best_index == last_finite_idx
                    and not best_index == len(results) - 1):
                # We have non-converged models on the upper bound of the
                # grid, we need to refine the grid there
                lam_1 = results[best_index][0]
                lam_0 = results[best_index + 1][0]
            
            elif best_index == len(results) - 1:
                lam_1 = results[best_index][0]
                lam_0 = 0.01 * results[best_index][0]
            
            else:
                lam_1 = results[best_index - 1][0]
                lam_0 = results[best_index + 1][0]

            if not isinstance(self.lams, collections.Sequence):
                path = np.logspace(np.log10(lam_1), np.log10(lam_0),
                                     self.lams + 2)
                path = path[1:-1]

            if self.verbose and n_refinements > 1:
                print('[GraphLassoCV] Done refinement % 2i out of %i: % 3is'
                      % (rr + 1, n_refinements, time.time() - t0))

        results = list(zip(*results))
        grid_scores = list(results[1])
        lams = list(results[0])
        
        # Finally, compute the score with lambda = 0
        lams.append(0)
        grid_scores.append(cross_val_score(EmpiricalCovariance(), X,
                                           cv=cv, n_jobs=self.n_jobs))
        self.grid_scores = np.array(grid_scores)
        self.lam_ = self.lam * lams[best_index]
        self.cv_lams_ = [self.lam * l for l in lams]

        # Finally fit the model with the selected lambda
        if self.method == 'quic':
            (self.precision_, self.covariance_, self.opt_, self.cputime_, 
            self.iters_, self.duality_gap_) = quic(self.sample_covariance_,
                                                self.lam_,
                                                mode='default',
                                                tol=self.tol,
                                                max_iter=self.max_iter,
                                                Theta0=self.Theta0,
                                                Sigma0=self.Sigma0,
                                                path=None,
                                                msg=self.verbose)
        else:
            raise NotImplementedError(
                "Only method='quic' has been implemented.")

        self.is_fitted = True
        return self


class QuicGraphLassoEBIC(InverseCovarianceEstimator):
    """
    Computes a sparse inverse covariance matrix estimation using quadratic 
    approximation and EBIC model selection. (Convenience Class)

    Note: This estimate can be obtained using the more general QuicGraphLasso
          estimator and taking advantage of `ebic_select()` and 
          `lambda_at_index()` methods.
    
    See analogous sklearn.linear_model.LassoLarsIC.

    Parameters
    -----------        
    lam : scalar or 2D ndarray, shape (n_features, n_features) (default=0.5)
        Regularization parameters per element of the inverse covariance matrix.
    
        If a scalar lambda is used, a penalty matrix will be generated 
        containing lambda for all values in both upper and lower triangles
        and zeros along the diagonal.  This differs from the scalar graphical 
        lasso by the diagonal. To replicate the scalar formulation you must 
        manualy pass in lam * np.ones((n_features, n_features)).

    path : array of floats or int (default=100)
        An array of float values for scaling lam.
        An int will choose the number of log-scale points to fit.

    gamma : float (default=0)
        Extended Bayesian Information Criteria (EBIC) for model selection.
        Choice of gamma=0 leads to classical BIC
        Positive gamma leads to stronger penalization of large graphs.

    tol : float (default=1e-6)
        Convergence threshold.

    max_iter : int (default=1000)
        Maximum number of Newton iterations.

    Theta0 : 2D ndarray, shape (n_features, n_features) (default=None) 
        Initial guess for the inverse covariance matrix. If not provided, the 
        diagonal identity matrix is used.

    Sigma0 : 2D ndarray, shape (n_features, n_features) (default=None)
        Initial guess for the covariance matrix. If not provided the diagonal 
        identity matrix is used.

    method : one of 'quic'... (default=quic)

    verbose : integer
        Used in quic routine.

    init_method : one of 'corrcoef', 'cov', or custom function
        Computes initial covariance and scales lambda appropriately.
        Custom function must return ((n_features, n_features) ndarray, float)
        where the scalar parameter will be used to scale the penalty lam.

    auto_scale : bool
        If True, will compute self.lam_scale_ = max off-diagonal value when 
        init_method='cov'. 
        If false, then self.lam_scale_ = 1.
        lam_scale_ is used to scale user-supplied self.lam during fit.

    Attributes
    ----------
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.

    sample_covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated sample covariance matrix

    lam_ : (float)
        Lambda chosen by EBIC (with scaling already applied).
    """
    def __init__(self, lam=1.0, path=100, gamma=0, tol=1e-6, max_iter=1000,
                 Theta0=None, Sigma0=None,  method='quic', verbose=0,
                 score_metric='log_likelihood', init_method='corrcoef',
                 auto_scale=True):
        # quic-specific params
        self.lam = lam
        self.tol = tol
        self.max_iter = max_iter
        self.Theta0 = Theta0
        self.Sigma0 = Sigma0
        self.method = method
        self.verbose = verbose
        self.path = path
        self.gamma = gamma

        # quic-specific outputs
        self.opt_ = None
        self.cputime_ = None
        self.iters_ = None
        self.duality_gap_ = None

        # these must be updated upon self.fit()
        self.path_ = None
        self.sample_covariance_ = None
        self.lam_scale_ = None
        self.lam_ = None
        self.is_fitted = False

        super(QuicGraphLassoEBIC, self).__init__(
                init_method=init_method,
                score_metric=score_metric,
                auto_scale=auto_scale)

    def fit(self, X, y=None, **fit_params):
        """Fits the inverse covariance model according to the given training 
        data and parameters.

        Parameters
        -----------
        X : 2D ndarray, shape (n_features, n_features)
            Input data.

        Returns
        -------
        self
        """
        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)
        self.init_coefs(X)

        # either use passed in path, or make our own path
        lam_1 = self.lam_scale_
        lam_0 = 1e-2 * lam_1
        if self.path is None:
            self.path_ = np.logspace(np.log10(lam_0), np.log10(lam_1), 100)[::-1]
        elif isinstance(self.path, int):
            self.path_ = np.logspace(
                    np.log10(lam_0), np.log10(lam_1), self.path)[::-1]
        else:
            self.path_ = self.path

        self.path_ = _validate_path(self.path_)

        # fit along the path, temporarily populate 
        # self.precision_, self.covariance_ with path values so we can use our
        # inherited selection function
        if self.method == 'quic':
            (self.precision_, self.covariance_, _, _, 
            _, _) = quic(self.sample_covariance_,
                        self.lam * self.lam_scale_,
                        mode='path',
                        tol=self.tol,
                        max_iter=self.max_iter,
                        Theta0=self.Theta0,
                        Sigma0=self.Sigma0,
                        path=self.path_,
                        msg=self.verbose)
            self.is_fitted = True
        else:
            raise NotImplementedError(
                "Only method='quic' has been implemented.")

        # apply EBIC criteria
        best_lam_idx = self.ebic_select(gamma=self.gamma)
        self.lam_ = self.lam * self.lam_scale_ * self.path_[best_lam_idx]
        self.precision_ = self.precision_[best_lam_idx]
        self.covariance_ = self.covariance_[best_lam_idx]

        self.is_fitted = True
        return self



import numpy as np
from sklearn.base import BaseEstimator, clone
from sklearn.utils import check_array, as_float_array

from .inverse_covariance import _init_coefs
from . import QuicGraphLassoCV


def _check_psd(m):
    return np.all(np.linalg.eigvals(m) >= 0)


def _fully_random_weights(n_features, lam_scale):
    """Generate a symmetric random matrix with zeros along the diagonal.
    """
    weights = np.zeros((n_features, n_features))
    n_off_diag = (n_features ** 2 - n_features) / 2 
    weights[np.triu_indices(n_features, k=1)] =\
            0.1 * lam_scale * np.random.randn(n_off_diag) + (0.25 * lam_scale)
    weights[weights < 0] = 0
    weights = weights + weights.T
    return weights


def _random_weights(n_features, lam, lam_perturb):
    """Generate a symmetric random matrix with zeros along the diagnoal and 
    non-zero elements take the value {lam * lam_perturb, lam / lam_perturb} 
    with probability 1/2.
    """
    weights = np.zeros((n_features, n_features))
    n_off_diag = (n_features ** 2 - n_features) / 2
    berns = np.random.binomial(1, 0.5, size=n_off_diag)
    vals = np.zeros(berns.shape)
    vals[berns == 0] = 1. * lam * lam_perturb
    vals[berns == 1] = 1. * lam / lam_perturb
    weights[np.triu_indices(n_features, k=1)] = vals
    weights[weights < 0] = 0
    weights = weights + weights.T
    return weights


def _fix_weights(weight_fun, *args):
    """Ensure random weight matrix is valid.

    TODO:  The diagonally dominant tuning currently doesn't make sense.  
           Our weight matrix has zeros along the diagonal, so multiplying by 
           a diagonal matrix results in a zero-matrix.
    """
    weights = weight_fun(*args)

    # TODO: fix this
    # disable checks for now
    return weights

    # if positive semidefinite, then we're good as is
    if _check_psd(weights):
        return weights

    # make diagonally dominant
    off_diag_sums = np.sum(weights, axis=1) # NOTE: assumes diag is zero
    mod_mat = np.linalg.inv(np.sqrt(np.diag(off_diag_sums)))
    return np.dot(mod_mat, weights, mod_mat)


def _default_bootstrap(n_samples, num_subsamples):
    """Returns an array of integers (0, n_samples-1)^num_subsamples.
    """
    return np.random.permutation(n_samples)[:num_subsamples]


class ModelAverage(BaseEstimator):
    """
    Randomized model averaging meta-estimator.

    See analogous sklearn.linear_model.BaseRandomizedLinearModel.
    
    Parameters
    -----------        
    estimator : An inverse covariance estimator instance
        After being fit, estimator.precision_ must either be a matrix with the 
        precision or a list of precision matrices (e.g., path mode).

    n_trials : int (default=100)
        Number of random subsets for which to bootstrap the data.

    subsample : float in range (0, 1) (default=0.3)
        Fraction of examples to subsample in each bootstrap trial.

    normalize : bool (default=True)
        Determines whether the proportion_ matrix should be normalized to have
        values in the range (0, 1) or should be absolute.

    lam : float (default=0.5)
        Scalar lambda penalty used in penalization='random' mode.  Will be 
        ignored in all other modes.

    lam_perturb : float \in (0, 1) (default=0.5)
        Scalar perturbation parameter used in penalization='random'.  Will be 
        ignored in all other modes.

    penalization : one of 'subsampling', 'random' (default), 'fully-random' 
        Strategy for generating new random penalization in each trial.
        
        subsampling: Only the observations will be subsampled, the original
                     penalty supplied in the estimator instance will be used.
                     Use this technique when the estimator does not support 
                     matrix penalization (e.g., sklearn GraphLasso).

        random: In addition to randomly subsampling the observations, 'random'
                applies a randomly-perturbed 'lam' weight matrix.  The entries 
                of the matrix take the value 
                {lam * lam_perturb, lam / lam_perturb} with probability 1/2. 
                User must supply a scalar 'lam' and 'lam_perturb' parameters. 

        fully-random: In addition to randomly subsampling the observations,
                      'fully-random' generates a symmetric Gaussian matrix 
                      appropriately scaled for the data. 

        For more information on 'random' penalization, see:
            "Stability Selection"
            N. Meinhausen and P. Buehlmann, May 2009

            "Random lasso" 
            S. Wang, B. Nan, S. Rosset, and J. Zhu, Apr 2011

        For more information on 'fully-random', see:
            "Mixed effects models for resampled network statistics improves
            statistical power to find differences in multi-subject functional
            connectivity" 
            M. Narayan and G. Allen, March 2016

    support_thresh : float (0, 1)
        Threshold for estimating supports from proportions.  This is provided
        for convience.

    use_cache : bool (default=True)
        If false, will optionally not cache each estimator instance and 
        penalization instance (to save memory).

    penalty_name : string (default='lam')
        Name of the penalty kwarg in the estimator.  This parameter is 
        unimportant if penalization='subsampling'.

    bootstrap : callable fun (default=_default_bootstrap)
        A function that takes n_samples, num_subsamples as inputs and returns 
        a list of sample indices in the range (0, n_samples-1).
        By default, indices are uniformly subsampled.

    Attributes
    ----------
    proportion_ : matrix of size (n_features, n_features)
        Each entry indicates the sample probability (or count) of whether the 
        inverse covariance is non-zero.

    support_ : matrix of size (n_features, n_features)
        Support estimate via thresholding proportions by support_thresh.

    estimators_ : list of estimator instances (n_trials, )
        The estimator instance from each trial.  
        This returns an empty list if use_cache=False.

    lams_ : list of penalization matrices (n_trials, )
        The penalization matrix chosen in each trial.
        This returns an empty list if penalization='subsampling'.
    
    subsets_ : list of subset indices (n_trials, )
        The example indices chosen in each trial.
        This returns an empty list if use_cache=False.

    lam_ : float
        Average matrix value used among lam_ for all estimators.
    """
    def __init__(self, estimator=None, n_trials=100, subsample=0.3, 
                 normalize=True, lam=0.5, lam_perturb=0.5, penalization='random',
                 use_cache=True, penalty_name='lam', support_thresh=0.5,
                 bootstrap=_default_bootstrap):
        self.estimator = estimator 
        self.n_trials = n_trials
        self.subsample = subsample
        self.normalize = normalize
        self.lam = lam 
        self.lam_perturb = lam_perturb
        self.penalization = penalization
        self.use_cache = use_cache
        self.penalty_name = penalty_name
        self.support_thresh = support_thresh
        self.bootstrap = bootstrap

        self.proportion_ = None
        self.support_ = None
        self.lam_ = None
        self.lam_scale_ = None
        self.estimators_ = []
        self.lams_ = []
        self.subsets_ = []

        # default to QuicGraphLassoCV
        if self.estimator is None:
            self.estimator = QuicGraphLassoCV()

        if self.penalization != 'subsampling' and\
                not hasattr(self.estimator, self.penalty_name):
            raise ValueError(("Must specify valid penalty for "
                              "estimator: {}.".format(self.penalty_name)))

    def fit(self, X, y=None):
        """Learn a model averaged proportion matrix for X.
        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the proportion matrix.
        """
        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)

        n_samples, n_features = X.shape
        _, self.lam_scale_ = _init_coefs(X, method='cov')
        
        self.lam_ = 0.0
        self.proportion_ = np.zeros((n_features, n_features))
        for nn in range(self.n_trials):
            prec_is_real = False
            while not prec_is_real:
                lam = None
                if self.penalization == 'subsampling':
                    pass
                elif self.penalization == 'random':
                    lam = _fix_weights(_random_weights, 
                                       n_features,
                                       self.lam,
                                       self.lam_perturb)
                elif self.penalization == 'fully-random':
                    lam = _fix_weights(_fully_random_weights,
                                       n_features,
                                       self.lam_scale_)
                else:
                    raise NotImplementedError(
                            ("Only penalization = 'subsampling', "
                            "'random', and 'fully-random' have "
                            "been implemented."))

                # new instance of estimator
                new_estimator = clone(self.estimator)
                if lam is not None:
                    new_estimator.set_params(**{
                        self.penalty_name: lam,
                    }) 

                # fit estimator
                num_subsamples = int(self.subsample * n_samples)
                rp = self.bootstrap(n_samples, num_subsamples)
                new_estimator.fit(X[rp, :])

                # check that new_estimator.precision_ is real
                # if not, skip this lam and try again
                if isinstance(new_estimator.precision_, list):
                    prec_real_bools = []
                    for prec in new_estimator.precision_:
                        prec_real_bools.append(np.all(np.isreal(prec)))

                    prec_is_real = np.all(np.array(prec_real_bools) == True)
                
                elif isinstance(new_estimator.precision_, np.ndarray):
                    prec_is_real = np.all(np.isreal(new_estimator.precision_))

                else:
                    raise ValueError("Estimator returned invalid precision_.")

            # update proportions
            if isinstance(new_estimator.precision_, list):
                for prec in new_estimator.precision_:
                    self.proportion_[np.nonzero(prec)] += 1.
            
            elif isinstance(new_estimator.precision_, np.ndarray):
                self.proportion_[np.nonzero(new_estimator.precision_)] += 1.
            
            else:
                raise ValueError("Estimator returned invalid precision_.")

            # estimate support locations
            threshold = self.support_thresh * self.n_trials
            self.support_ = np.zeros(self.proportion_.shape)
            self.support_[self.proportion_ > threshold] = 1.0

            # currently, dont estimate self.lam_ if penalty_name is different
            if self.penalty_name == 'lam':
                self.lam_ += np.mean(new_estimator.lam_.flat)
            
            # save estimators, subsets, and lambdas
            if self.use_cache:
                self.estimators_.append(new_estimator)
                self.subsets_.append(rp)
                if lam is not None:
                    self.lams_.append(lam)

        self.lam_ /= self.n_trials
        if self.normalize:
            self.proportion_ /= self.n_trials

    @property
    def precision_(self):
        '''Convenience property to make compatible with AdaptiveGraphLasso.
        This is not a very good precision estimate.
        '''
        return self.support_

    @property
    def covariance_(self):
        '''Convenience property to make compatible with AdaptiveGraphLasso.
        This is not a very good covariance estimate.
        '''
        return np.linalg.inv(self.support_)
    



import numpy as np 

from sklearn.base import clone
from sklearn.datasets import make_sparse_spd_matrix
from sklearn.externals.joblib import Parallel, delayed
from matplotlib import pyplot as plt
#import seaborn

from .. import QuicGraphLasso


plt.ion()
prng = np.random.RandomState(1)



def _new_graph(n_features, alpha):
    global prng
    prec = make_sparse_spd_matrix(n_features,
                                  alpha=alpha, # prob that a coeff is zero
                                  smallest_coef=0.7,
                                  largest_coef=0.7,
                                  random_state=prng)
    cov = np.linalg.inv(prec)
    d = np.sqrt(np.diag(cov))
    cov /= d
    cov /= d[:, np.newaxis]
    prec *= d
    prec *= d[:, np.newaxis]
    return cov, prec


def _new_sample(n_samples, n_features, cov):
    X = prng.multivariate_normal(np.zeros(n_features), cov, size=n_samples)
    X -= X.mean(axis=0)
    X /= X.std(axis=0)
    return X


def _plot_avg_error(results, grid, ks):
    plt.figure()
    plt.plot(grid, results.T, lw=2)
    plt.xlabel('n/p (n_samples / n_features)')
    plt.ylabel('Average error')
    legend_text = []
    for ks in ks:
        legend_text.append('sparsity={}'.format(ks))
    plt.legend(legend_text)
    plt.show()


def _support_diff(m, m_hat):
    '''Count the number of different elements in the support in one triangle,
    not including the diagonal. 
    '''
    n_features, _ = m.shape

    m_no_diag = m.copy()
    m_no_diag[np.diag_indices(n_features)] = 0
    m_hat_no_diag = m_hat.copy()
    m_hat_no_diag[np.diag_indices(n_features)] = 0

    m_nnz = len(np.nonzero(m_no_diag.flat)[0])
    m_hat_nnz = len(np.nonzero(m_hat_no_diag.flat)[0])

    nnz_intersect = len(np.intersect1d(np.nonzero(m_no_diag.flat)[0],
                                       np.nonzero(m_hat_no_diag.flat)[0]))
    return (m_nnz + m_hat_nnz - (2 * nnz_intersect)) / 2.0


def _false_support(m, m_hat):
    '''Count the number of false positive and false negatives supports in 
    m_hat in one triangle, not including the diagonal.
    '''
    n_features, _ = m.shape

    m_no_diag = m.copy()
    m_no_diag[np.diag_indices(n_features)] = 0
    m_hat_no_diag = m_hat.copy()
    m_hat_no_diag[np.diag_indices(n_features)] = 0

    m_nnz = len(np.nonzero(m_no_diag.flat)[0])
    m_hat_nnz = len(np.nonzero(m_hat_no_diag.flat)[0])

    nnz_intersect = len(np.intersect1d(np.nonzero(m_no_diag.flat)[0],
                                       np.nonzero(m_hat_no_diag.flat)[0]))

    false_positives = (m_hat_nnz - nnz_intersect) / 2.0
    false_negatives = (m_nnz - nnz_intersect) / 2.0
    return false_positives, false_negatives


def _ae_trial(trial_estimator, n_samples, n_features, cov, prec):
    X = _new_sample(n_samples, n_features, cov)
    new_estimator = clone(trial_estimator)
    new_estimator.fit(X)

    error_fro = np.linalg.norm(prec - new_estimator.precision_, ord='fro')
    error_supp = _support_diff(prec, new_estimator.precision_)
    error_fp, error_fn = _false_support(prec, new_estimator.precision_)

    return error_fro, error_supp, error_fp, error_fn


class AverageError(object):
    """Compute the average error of a model selector for
    different values of alpha over grid of n_samples / n_features.

    Use this to compare model selection methods.

    For each choice of alpha, we select a fixed test graph.
    For each choice of n_samples / n_features, we learn the model selection
    penalty just once and apply this learned value to each subsequent random
    trial (new instances with the fixed covariance).

    Parameters
    -----------        
    model_selection_estimator : An inverse covariance estimator instance 
        This estimator must be able to select a penalization parameter. 
        Use .penalty_ to obtain selected penalty.

    n_features : int (default=50)
        Fixed number of features to test.

    n_trials : int (default=100)
        Number of examples to draw to measure P(recovery).

    penalty_ : string (default='lam_')
        Name of the selected best penalty in estimator
        e.g., 'lam_' for QuicGraphLassoCV, QuicGraphLassoEBIC,
              'alpha_' for GraphLassoCV

    n_grid_points : int (default=10)
        Number of grid points for sampling n_samples / n_features between (0,1)

    verbose : bool (default=False)
        Print out progress information.

    n_jobs: int, optional
        number of jobs to run in parallel (default 1).

    Methods
    ----------
    show() : Plot the results.

    Attributes
    ----------
    grid_ : array of size (n_grid_points, )
        Array of n_samples / n_features ratios.

    alphas_ : array of size (n_alpha_grid_points, )
        Array of alphas used to generate test graphs 
        (see .statistical_power._new_graph)

    ks_ : array of size (n_alpha_grid_points, )
        The sparsity of each test graph.

    error_fro_ : matrix of size (n_alpha_grid_points, n_grid_points)
        The average Frobenius error for each alpha and 
        n_samples / n_features grid point.

    error_supp_ : matrix of size (n_alpha_grid_points, n_grid_points)
        The average support difference for each alpha and 
        n_samples / n_features grid point.

    error_fp_ : matrix of size (n_alpha_grid_points, n_grid_points)
        The average false positive difference for each alpha and 
        n_samples / n_features grid point.

    error_fn_ : matrix of size (n_alpha_grid_points, n_grid_points)
        The average false negative difference for each alpha and 
        n_samples / n_features grid point.
    """
    def __init__(self, model_selection_estimator=None, n_features=50, 
                n_trials=100, n_grid_points=10, verbose=False, penalty_='lam_',
                n_jobs=1):
        self.model_selection_estimator = model_selection_estimator  
        self.n_features = n_features
        self.n_grid_points = n_grid_points
        self.n_trials = n_trials
        self.verbose = verbose
        self.penalty_ = penalty_ # class name for model selected penalty
        self.n_jobs = n_jobs

        self.is_fitted = False
        self.error_fro_ = None
        self.error_supp_ = None
        self.error_fp_ = None
        self.error_fn_ = None
        self.alphas_ = None
        self.ks_ = None
        self.grid_ = None
 
    def fit(self, X=None, y=None):
        n_alpha_grid_points = 3

        self.error_fro_ = np.zeros((n_alpha_grid_points, self.n_grid_points))
        self.error_supp_ = np.zeros((n_alpha_grid_points, self.n_grid_points))
        self.error_fp_ = np.zeros((n_alpha_grid_points, self.n_grid_points))
        self.error_fn_ = np.zeros((n_alpha_grid_points, self.n_grid_points))

        self.grid_ = np.linspace(1, 4, self.n_grid_points)
        self.alphas_ = np.linspace(0.95, 0.99, n_alpha_grid_points)[::-1]
        self.ks_ = []

        for aidx, alpha in enumerate(self.alphas_):
            if self.verbose:
                print 'at alpha {} ({}/{})'.format(
                    alpha,
                    aidx,
                    n_alpha_grid_points,
                )

            # draw a new fixed graph for alpha
            cov, prec = _new_graph(self.n_features, alpha)
            n_nonzero_prec = np.count_nonzero(prec.flat)
            self.ks_.append(n_nonzero_prec)
            if self.verbose:
                print '   Graph has {} nonzero entries'.format(n_nonzero_prec)

            for sidx, sample_grid in enumerate(self.grid_):
                n_samples = int(sample_grid * self.n_features)
                
                # model selection (once)
                X = _new_sample(n_samples, self.n_features, cov)
                ms_estimator = clone(self.model_selection_estimator)
                ms_estimator.fit(X)
                lam = getattr(ms_estimator, self.penalty_)
                
                if self.verbose:
                    display_lam = lam
                    if isinstance(lam, np.ndarray):
                        display_lam = np.linalg.norm(lam)
                    print '   ({}/{}), n_samples = {}, selected lambda = {}'.format(
                            sidx,
                            self.n_grid_points,
                            n_samples,
                            display_lam)

                # setup default trial estimator
                trial_estimator = QuicGraphLasso(lam=lam,
                                                 mode='default',
                                                 init_method='corrcoef')

                # estimate statistical power
                errors = Parallel(
                    n_jobs=self.n_jobs,
                    verbose=False,
                    backend='threading',
                    #max_nbytes=None,
                    #batch_size=1,
                )(
                    delayed(_ae_trial)(
                        trial_estimator, n_samples, self.n_features, cov, prec
                    )
                    for nn in range(self.n_trials))

                error_fro, error_supp, error_fp, error_fn = zip(*errors)
                self.error_fro_[aidx, sidx] = np.mean(error_fro)
                self.error_supp_[aidx, sidx] = np.mean(error_supp)
                self.error_fp_[aidx, sidx] = np.mean(error_fp)
                self.error_fn_[aidx, sidx] = np.mean(error_fn)

            if self.verbose:
                print 'Results at this row:'
                print '   fro = {}'.format(self.error_fro_[aidx, :])
                print '   supp = {}'.format(self.error_supp_[aidx, :])
                print '   fp = {}'.format(self.error_fp_[aidx, :])
                print '   fn = {}'.format(self.error_fn_[aidx, :])

        self.is_fitted = True
        return self

    def show(self):
        if not self.is_fitted:
            print 'Not fitted.'
            return

        errors_to_plot = [
            ('Frobenius', self.error_fro_),
            ('Support', self.error_supp_),
            ('False Positive', self.error_fp_),
            ('False Negative', self.error_fn_),
        ]
        for name, result in errors_to_plot:
            _plot_avg_error(result, self.grid_, self.ks_)
            plt.title(name)




from .inverse_covariance import (
    InverseCovarianceEstimator,
)
from .quic_graph_lasso import (
    quic,
    QuicGraphLasso,
    QuicGraphLassoCV,
    QuicGraphLassoEBIC,
)
from .metrics import (
    log_likelihood,
    kl_loss,
    quadratic_loss,
    ebic,
)
from .model_average import ModelAverage
from .adaptive_graph_lasso import AdaptiveGraphLasso

__all__ = [
    'InverseCovarianceEstimator',
    'quic',
    'QuicGraphLasso',
    'QuicGraphLassoCV',
    'QuicGraphLassoEBIC',
    'log_likelihood',
    'kl_loss',
    'quadratic_loss',
    'ebic',
    'ModelAverage',
    'AdaptiveGraphLasso',
]


import numpy
from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import platform


if platform.system() == 'Darwin':
    extra_compile_args = ['-I/System/Library/Frameworks/vecLib.framework/Headers']
    if 'ppc' in platform.machine():
        extra_compile_args.append('-faltivec')

    extra_link_args = ["-Wl,-framework", "-Wl,Accelerate"]
    include_dirs = [numpy.get_include()]
        
else:
    include_dirs = [numpy.get_include(), "/usr/local/include"]
    extra_compile_args=['-msse2', '-O2', '-fPIC', '-w']
    extra_link_args=["-llapack"]



ext_module = Extension(
    name="pyquic",
    sources=["QUIC.C", "pyquic.pyx"],
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c++"
)


if __name__ == "__main__":
    setup(
        name = "pyquic",
        cmdclass = {"build_ext": build_ext},
        ext_modules = [ext_module]
    )



import numpy as np
from sklearn.base import BaseEstimator 

from . import metrics


def _init_coefs(X, method='corrcoef'):
    if method == 'corrcoef':
        return np.corrcoef(X, rowvar=False), 1.0
    elif method == 'cov':   
        init_cov = np.cov(X, rowvar=False)
        return init_cov, np.max(np.abs(np.triu(init_cov)))
    elif callable(method):
        return method(X)
    else:
        raise ValueError(("initialize_method must be 'corrcoef' or 'cov', "
            "passed \'{}\' .".format(method)))


def _compute_error(comp_cov, covariance_, precision_, score_metric='frobenius'):
    """Computes the covariance error vs. comp_cov.
        
    Parameters
    ----------
    comp_cov : array-like, shape = (n_features, n_features)
        The precision to compare with.
        This should normally be the test sample covariance/precision.
            
    scaling : bool
        If True, the squared error norm is divided by n_features.
        If False (default), the squared error norm is not rescaled.

    score_metric : str
        The type of norm used to compute the error between the estimated 
        self.precision, self.covariance and the reference `comp_cov`. 
        Available error types:
        
        - 'frobenius' (default): sqrt(tr(A^t.A))
        - 'spectral': sqrt(max(eigenvalues(A^t.A))
        - 'kl': kl-divergence 
        - 'quadratic': quadratic loss
        - 'log_likelihood': negative log likelihood
    
    squared : bool
        Whether to compute the squared error norm or the error norm.
        If True (default), the squared error norm is returned.
        If False, the error norm is returned.
    """
    if score_metric == "frobenius":
        error = comp_cov - covariance_
        return np.sum(error ** 2)                        
    elif score_metric == "spectral":
        error = comp_cov - covariance_
        return np.amax(np.linalg.svdvals(np.dot(error.T, error)))
    elif score_metric == "kl":
        return metrics.kl_loss(comp_cov, precision_)
    elif score_metric == "quadratic":
        return metrics.quadratic_loss(comp_cov, precision_)
    elif score_metric == "log_likelihood":
        return -metrics.log_likelihood(comp_cov, precision_)
    else:
        raise NotImplementedError(("Must be frobenius, spectral, kl, "
                                   "quadratic, or log_likelihood"))


def _validate_path(path):
    """Sorts path values from largest to smallest.

    Will warn if path parameter was not already sorted.
    """
    if path is None:
        return None

    new_path = np.array(sorted(set(path), reverse=True))
    if new_path[0] != path[0]:
        print 'Warning: Path must be sorted largest to smallest.'

    return new_path

    
class InverseCovarianceEstimator(BaseEstimator):
    """
    Base class for inverse covariance estimators.

    Provides initialization method, metrics, scoring function, 
    and ebic model selection.

    Parameters
    -----------            
    score_metric : one of 'log_likelihood' (default), 'frobenius', 'spectral',
                  'kl', or 'quadratic'
        Used for computing self.score().

    init_method : one of 'corrcoef', 'cov', or custom function
        Computes initial covariance and scales lambda appropriately.
        Custom function must return ((n_features, n_features) ndarray, float)
        where the scalar parameter will be used to scale the penalty lam.

    auto_scale : bool
        If True, will compute self.lam_scale_ = max off-diagonal value when 
        init_method='cov'. 
        If false, then self.lam_scale_ = 1.
        lam_scale_ is used to scale user-supplied self.lam during fit.

    Attributes
    ----------
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix
        
        This can also be a len(path) list of
        2D ndarray, shape (n_features, n_features)
        (e.g., see mode='path' in QuicGraphLasso)

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.
        
        This can also be a len(path) list of
        2D ndarray, shape (n_features, n_features)
        (e.g., see mode='path' in QuicGraphLasso)

    sample_covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated sample covariance matrix

    lam_scale_ : (float)
        Additional scaling factor on lambda (due to magnitude of 
        sample_covariance_ values).
    """
    def __init__(self, score_metric='log_likelihood', init_method='cov',
                 auto_scale=True):
        self.score_metric = score_metric
        self.init_method = init_method
        self.auto_scale = auto_scale

        self.covariance_ = None  # assumes a matrix of a list of matrices
        self.precision_ = None  # assumes a matrix of a list of matrices

        # these must be updated upon self.fit()
        # the first 4 will be set if self.init_coefs is used.
        #   self.sample_covariance_
        #   self.lam_scale_
        #   self.n_samples
        #   self.n_features
        self.is_fitted = False 

        super(InverseCovarianceEstimator, self).__init__()


    def init_coefs(self, X):
        """Computes ...

        Initialize the following values:
            self.n_samples
            self.n_features
            self.sample_covariance_
            self.lam_scale_
        """
        self.n_samples, self.n_features = X.shape
        self.sample_covariance_, self.lam_scale_ = _init_coefs(
                X,
                method=self.init_method)

        if not self.auto_scale:
            self.lam_scale_ = 1.0


    def score(self, X_test, y=None):
        """Computes the score between cov/prec of sample covariance of X_test 
        and X via 'score_metric'.

        Note: We want to maximize score so we return the negative error.
       
        Parameters
        ----------
        X_test : array-like, shape = [n_samples, n_features]
            Test data of which we compute the likelihood, where n_samples is
            the number of samples and n_features is the number of features.
            X_test is assumed to be drawn from the same distribution than
            the data used in fit (including centering).
        
        y : not used.
        
        Returns
        -------
        result : float or list of floats
            The negative of the min error between `self.covariance_` and
            the sample covariance of X_test.
        """        
        if isinstance(self.precision_, list):
            print 'Warning: returning a list of scores.'

        S_test, lam_scale_test = _init_coefs(
                X_test,
                method=self.init_method)
        error = self.cov_error(S_test, score_metric=self.score_metric)

        # maximize score with -error
        return -error


    def cov_error(self, comp_cov, score_metric='frobenius'):
        """Computes the covariance error vs. comp_cov.

        May require self.path_
        
        Parameters
        ----------
        comp_cov : array-like, shape = (n_features, n_features)
            The precision to compare with.
            This should normally be the test sample covariance/precision.
                
        scaling : bool
            If True, the squared error norm is divided by n_features.
            If False (default), the squared error norm is not rescaled.

        score_metric : str
            The type of norm used to compute the error between the estimated 
            self.precision, self.covariance and the reference `comp_cov`. 
            Available error types:
            
            - 'frobenius' (default): sqrt(tr(A^t.A))
            - 'spectral': sqrt(max(eigenvalues(A^t.A))
            - 'kl': kl-divergence 
            - 'quadratic': quadratic loss
            - 'log_likelihood': negative log likelihood
        
        squared : bool
            Whether to compute the squared error norm or the error norm.
            If True (default), the squared error norm is returned.
            If False, the error norm is returned.
        
        Returns
        -------
        The min error between `self.covariance_` and `comp_cov`.
        
        If self.precision_ is a list, returns errors for each matrix, otherwise
        returns a scalar.
        """  
        if not isinstance(self.precision_, list):
            return _compute_error(comp_cov,
                                self.covariance_,
                                self.precision_,
                                score_metric)

        path_errors = []
        for lidx, lam in enumerate(self.path_):
            path_errors.append(_compute_error(comp_cov,
                                            self.covariance_[lidx],
                                            self.precision_[lidx],
                                            score_metric))

        return np.array(path_errors)


    def ebic(self, gamma=0):
        """Compute EBIC scores for each model. If model is not "path" then 
        returns a scalar score value. 

        May require self.path_

        See:
            Extended Bayesian Information Criteria for Gaussian Graphical Models
            R. Foygel and M. Drton
            NIPS 2010
    
        Parameters
        ----------
        gamma : (float) \in (0, 1)
            Choice of gamma=0 leads to classical BIC
            Positive gamma leads to stronger penalization of large graphs.

        Returns
        -------
        Scalar ebic score or list of ebic scores.
        """
        if not self.is_fitted:
            return

        if not isinstance(self.precision_, list):
            return metrics.ebic(self.sample_covariance_,
                                self.precision_,
                                self.n_samples,
                                self.n_features,
                                gamma=gamma)

        ebic_scores = []
        for lidx, lam in enumerate(self.path_):
            ebic_scores.append(metrics.ebic(
                    self.sample_covariance_,
                    self.precision_[lidx],
                    self.n_samples,
                    self.n_features,
                    gamma=gamma))

        return np.array(ebic_scores)


    def ebic_select(self, gamma=0):
        """Uses Extended Bayesian Information Criteria for model selection.

        Can only be used in path mode (doesn't really make sense otherwise).

        See:
            Extended Bayesian Information Criteria for Gaussian Graphical Models
            R. Foygel and M. Drton
            NIPS 2010

        Parameters
        ----------
        gamma : (float) \in (0, 1)
            Choice of gamma=0 leads to classical BIC
            Positive gamma leads to stronger penalization of large graphs.

        Returns
        -------
        Lambda index with best ebic score.  When multiple ebic scores are the 
        same, returns the smallest lambda (largest index) with minimum score.
        """
        if not isinstance(self.precision_, list):
            raise NotImplementedError("EBIC requires multiple models to select from.")
            return

        if not self.is_fitted:
            return

        ebic_scores = self.ebic(gamma=gamma)
        min_indices = np.where(np.abs(ebic_scores - ebic_scores.min()) < 1e-10) 
        return np.max(min_indices) 



"""Various utilities for Gaussian graphical models.
"""
import numpy as np
from sklearn.utils.testing import assert_array_equal
from matplotlib import pyplot as plt
import seaborn

plt.ion()


def _check_path(in_path):
    path = sorted(set(in_path), reverse=True)
    assert_array_equal(path, in_path)


def trace_plot(results, path):
    """Plot the change in precision (or covariance) coefficients as a function 
    of changing lambda and l1-norm.

    Parameters
    -----------
    results: array of len(path) of 2D ndarray, shape (n_features, n_features)
        This is either precision_ or covariance_ from an InverseCovariance
        estimator in path mode, or a list of results for individual runs of 
        the GraphLasso.

    path:  array of floats (descending)
        This is path of lambdas explored. 
    """
    _check_path(path)
    assert len(path) == len(results)
    assert len(results) > 0

    path = np.array(path)
    dim, _ = results[0].shape

    # reshape data a bit:  
    # flatten each matrix into a column (so that coeffs are examples)
    # compute l1-norm of each column
    l1_norms = []
    coeffs = np.zeros((dim**2, len(results)))
    for ridx, result in enumerate(results):
        coeffs[:, ridx] = result.flat 
        l1_norms.append(np.linalg.norm(coeffs[:, ridx]))

    plt.figure()

    # show coefficients as a function of lambda
    plt.subplot(1, 2, 1)
    for result in results:
        plt.plot(l1_norms, coeffs.T, lw=1)

    plt.xlim([np.min(l1_norms), np.max(l1_norms)])
    plt.ylabel('Coefficients')
    plt.xlabel('l1 Norm')

    # show coefficients as a function of lambda
    log_path = np.log(path)
    plt.subplot(1, 2, 2)
    for result in results:
        plt.plot(log_path, coeffs.T, lw=1)

    plt.xlim([np.min(log_path), np.max(log_path)])
    plt.ylabel('Coefficients')
    plt.xlabel('log-Lambda')

    plt.show()
    raw_input('Press any key to continue.')

    


import numpy as np 

from sklearn.base import clone
from sklearn.datasets import make_sparse_spd_matrix
from sklearn.externals.joblib import Parallel, delayed
from matplotlib import pyplot as plt
import seaborn

from .. import QuicGraphLasso


plt.ion()
prng = np.random.RandomState(1)



def _new_graph(n_features, alpha):
    global prng
    prec = make_sparse_spd_matrix(n_features,
                                  alpha=alpha, # prob that a coeff is zero
                                  smallest_coef=0.7,
                                  largest_coef=0.7,
                                  random_state=prng)
    cov = np.linalg.inv(prec)
    d = np.sqrt(np.diag(cov))
    cov /= d
    cov /= d[:, np.newaxis]
    prec *= d
    prec *= d[:, np.newaxis]
    return cov, prec


def _new_sample(n_samples, n_features, cov):
    X = prng.multivariate_normal(np.zeros(n_features), cov, size=n_samples)
    X -= X.mean(axis=0)
    X /= X.std(axis=0)
    return X


def _plot_spower(results, grid, ks):
    plt.figure()
    plt.plot(grid, results.T, lw=2)
    plt.xlabel('n/p (n_samples / n_features)')
    plt.ylabel('P(exact recovery)')
    legend_text = []
    for ks in ks:
        legend_text.append('sparsity={}'.format(ks))
    plt.legend(legend_text)
    plt.show()


def _exact_support(prec, prec_hat):
    # Q: why do we need something like this?, and why must eps be so big?
    # Q: can we automatically determine what this threshold should be?
    eps = 0.2
    #eps = np.finfo(prec_hat.dtype).eps # too small
    prec_hat[np.abs(prec_hat) <= eps] = 0.0
    
    return np.array_equal(
            np.nonzero(prec.flat)[0],
            np.nonzero(prec_hat.flat)[0])


def _sp_trial(trial_estimator, n_samples, n_features, cov, prec):
    X = _new_sample(n_samples, n_features, cov)
    new_estimator = clone(trial_estimator)
    new_estimator.fit(X)
    return _exact_support(prec, new_estimator.precision_)


class StatisticalPower(object):
    """Compute the statistical power P(exact support) of a model selector for
    different values of alpha over grid of n_samples / n_features.

    For each choice of alpha, we select a fixed test graph.
    
    For each choice of n_samples / n_features, we learn the model selection
    penalty just once and apply this learned value to each subsequent random
    trial (new instances with the fixed covariance).

    Parameters
    -----------        
    model_selection_estimator : An inverse covariance estimator instance 
        This estimator must be able to select a penalization parameter. 
        Use .penalty_ to obtain selected penalty.

    n_features : int (default=50)
        Fixed number of features to test.

    n_trials : int (default=100)
        Number of examples to draw to measure P(recovery).

    trial_estimator : An inverse covariance estimator instance (default=None)
        Estimator to use on each instance after selecting a penalty lambda.
        If None, this will use QuicGraphLasso with lambda obtained with 
        model_selection_estimator.
        Use .penalty to set selected penalty.

    penalty_ : string (default='lam_')
        Name of the selected best penalty in estimator
        e.g., 'lam_' for QuicGraphLassoCV, QuicGraphLassoEBIC,
              'alpha_' for GraphLassoCV

    penalty : string (default='lam')
        Name of the penalty kwarg in the estimator.  
        e.g., 'lam' for QuicGraphLasso, 'alpha' for GraphLasso

    n_grid_points : int (default=10)
        Number of grid points for sampling n_samples / n_features between (0,1)

    verbose : bool (default=False)
        Print out progress information.

    n_jobs: int, optional
        number of jobs to run in parallel (default 1).

    Methods
    ----------
    show() : Plot the results.

    Attributes
    ----------
    grid_ : array of size (n_grid_points, )
        Array of n_samples / n_features ratios.

    alphas_ : array of size (n_alpha_grid_points, )
        Array of alphas used to generate test graphs 
        (see .statistical_power._new_graph)

    ks_ : array of size (n_alpha_grid_points, )
        The sparsity of each test graph.

    results_ : matrix of size (n_alpha_grid_points, n_grid_points)
        The statisical power, P(exact support recovery) for each alpha and 
        n_samples / n_features grid point.
    """
    def __init__(self, model_selection_estimator=None, n_features=50, 
                trial_estimator=None, n_trials=100, n_grid_points=10,
                verbose=False, penalty_='lam_', penalty='lam', n_jobs=1):
        self.model_selection_estimator = model_selection_estimator  
        self.trial_estimator = trial_estimator
        self.n_features = n_features
        self.n_grid_points = n_grid_points
        self.n_trials = n_trials
        self.verbose = verbose
        self.penalty_ = penalty_ # class name for model selected penalty
        self.penalty = penalty # class name for setting penalty
        self.n_jobs = n_jobs

        self.is_fitted = False
        self.results_ = None
        self.alphas_ = None
        self.ks_ = None
        self.grid_ = None
 
    def fit(self, X=None, y=None):
        n_alpha_grid_points = 5

        self.results_ = np.zeros((n_alpha_grid_points, self.n_grid_points))
        self.grid_ = np.linspace(0.25, 4, self.n_grid_points)
        self.alphas_ = np.linspace(0.99, 0.999, n_alpha_grid_points)[::-1]
        self.ks_ = []

        for aidx, alpha in enumerate(self.alphas_):
            if self.verbose:
                print 'at alpha {} ({}/{})'.format(
                    alpha,
                    aidx,
                    n_alpha_grid_points,
                )

            # draw a new fixed graph for alpha
            cov, prec = _new_graph(self.n_features, alpha)
            n_nonzero_prec = np.count_nonzero(prec.flat)
            self.ks_.append(n_nonzero_prec)
            if self.verbose:
                print '   Graph has {} nonzero entries'.format(n_nonzero_prec)

            for sidx, sample_grid in enumerate(self.grid_):
                n_samples = int(sample_grid * self.n_features)
                
                # model selection (once)
                X = _new_sample(n_samples, self.n_features, cov)
                ms_estimator = clone(self.model_selection_estimator)
                ms_estimator.fit(X)
                lam = getattr(ms_estimator, self.penalty_)
                
                if self.verbose:
                    display_lam = lam
                    if isinstance(lam, np.ndarray):
                        display_lam = np.linalg.norm(lam)
                    print '   ({}/{}), n_samples = {}, selected lambda = {}'.format(
                            sidx,
                            self.n_grid_points,
                            n_samples,
                            display_lam)

                # setup default trial estimator
                if self.trial_estimator is None:
                    trial_estimator = QuicGraphLasso(lam=lam,
                                                     mode='default',
                                                     init_method='corrcoef')
                else:
                    trial_estimator = self.trial_estimator

                # patch trial estimator with this lambda
                trial_estimator.set_params(**{
                    self.penalty: lam, 
                })

                # estimate statistical power
                exact_support_counts = Parallel(
                    n_jobs=self.n_jobs,
                    verbose=False,
                    backend='threading',
                    #max_nbytes=None,
                    #batch_size=1,
                )(
                    delayed(_sp_trial)(
                        trial_estimator, n_samples, self.n_features, cov, prec
                    )
                    for nn in range(self.n_trials))

                self.results_[aidx, sidx] = 1. * np.sum(exact_support_counts) / self.n_trials

            if self.verbose:
                print 'Results at this row: {}'.format(self.results_[aidx, :])

        self.is_fitted = True
        return self

    def show(self):
        if not self.is_fitted:
            print 'Not fitted.'
            return

        _plot_spower(self.results_, self.grid_, self.ks_)




import numpy as np
from sklearn.base import clone
from sklearn.utils import check_array, as_float_array

from . import (
    QuicGraphLasso,
    QuicGraphLassoCV,
    InverseCovarianceEstimator,
)


class AdaptiveGraphLasso(InverseCovarianceEstimator):
    """
    Two-stage adaptive estimator.

    a) Compute the first estimate with an estimator of your choice
       (self.estimator).  This estimate must produce a single precision_
       output.

    b) The resulting coefficients are used to generate adaptive weights and 
       the QuicGraphLasso is refit with these weights.

    See:
        "High dimensional covariance estimation based on Gaussian graphical
        models"
        S. Zhou, P. R{\"u}htimann, M. Xu, and P. B{\"u}hlmann

        "Relaxed Lasso"
        N. Meinshausen, December 2006.

    Parameters
    -----------        
    estimator : GraphLasso instance with model selection (default=QuicGraphLassoCV())
        After being fit, estimator.precision_ must either be a matrix.

    method : one of 'binary', 'inverse_squared', 'inverse' (default='binary')
        binary: non-zero where estimator was zero, 1 else (gelato, "relaxed lasso")
        inverse_squared: 1/|coefficient|^2  (glasso)
        inverse: 1/|coefficient|

    Attributes
    ----------
    estimator_ : QuicGraphLasso instance
        The final estimator refit with adaptive weights.

    lam_ : 2D ndarray, shape (n_features, n_features)
        Adaptive weight matrix generated.
    """
    def __init__(self, estimator=None, method='binary'):
        self.estimator = estimator 
        self.method = method

        self.lam_ = None
        self.estimator_ = None

        # default to QuicGraphLassoCV
        if self.estimator is None:
            self.estimator = QuicGraphLassoCV()


    def _binary_weights(self, estimator):
        n_features, _ = estimator.precision_.shape
        lam = np.zeros((n_features, n_features))
        lam[estimator.precision_ == 0] = 1
        lam[np.diag_indices(n_features)] = 0
        return lam
        

    def _inverse_squared_weights(self, estimator):
        n_features, _ = estimator.precision_.shape
        lam = np.zeros((n_features, n_features))
        mask = estimator.precision_ != 0
        lam[mask] = 1. / (np.abs(estimator.precision_[mask]) ** 2)
        mask_0 = estimator.precision_ == 0
        lam[mask_0] = np.max(lam[mask].flat) # non-zero in appropriate scale range
        lam[np.diag_indices(n_features)] = 0
        return lam


    def _inverse_weights(self, estimator):
        n_features, _ = estimator.precision_.shape
        lam = np.zeros((n_features, n_features))
        mask = estimator.precision_ != 0
        lam[mask] = 1. / np.abs(estimator.precision_[mask])
        mask_0 = estimator.precision_ == 0
        lam[mask_0] = np.max(lam[mask].flat) # non-zero in appropriate scale range
        lam[np.diag_indices(n_features)] = 0
        return lam


    def fit(self, X, y=None):
        """Estimate the precision using an adaptive maximum likelihood estimator.
        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the proportion matrix.
        """
        X = check_array(X, ensure_min_features=2, estimator=self)
        X = as_float_array(X, copy=False, force_all_finite=False)

        n_samples, n_features = X.shape

        # perform first estimate
        new_estimator = clone(self.estimator)
        new_estimator.fit(X)

        if self.method == 'binary':
            # generate weights
            self.lam_ = self._binary_weights(new_estimator)

            # perform second step adaptive estimate
            self.estimator_ = QuicGraphLasso(lam=self.lam_ * new_estimator.lam_, 
                                             mode='default',
                                             init_method='cov',
                                             auto_scale=False) 
            self.estimator_.fit(X)
        
        elif self.method == 'inverse_squared':
            self.lam_ = self._inverse_squared_weights(new_estimator)

            # perform second step adaptive estimate
            self.estimator_ = QuicGraphLassoCV(lam=self.lam_ * new_estimator.lam_,
                                               auto_scale=False) 
            self.estimator_.fit(X)
        
        elif self.method == 'inverse':
            self.lam_ = self._inverse_weights(new_estimator)

            # perform second step adaptive estimate
            self.estimator_ = QuicGraphLassoCV(lam=self.lam_ * new_estimator.lam_,
                                               auto_scale=False)
            self.estimator_.fit(X)

        else:
            raise NotImplementedError(("Only method='binary', 'inverse_squared', or",
                                       "'inverse' have been implemented."))

        self.is_fitted = True
        return self

